{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE Strict            #-}

-- Copyright (C) 2022, Maxim Likhachev, <envrm@yandex.ru>
--
-- This program is free software: you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free
-- Software Foundation, version 3.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program. If not, see <http://www.gnu.org/licenses/>.

module Main (main) where

import           Data.Aeson                           (decode', decodeFileStrict', encodeFile)
import qualified Data.ByteString.Char8                as BS
import qualified Data.ByteString.Lazy.Char8           as BL
import           Data.Char                            (toLower, toUpper)
import           Data.Map                             (Map, delete, elems, empty, filter, findWithDefault, insertWith,
                                                       mapWithKey, singleton)
import           Data.Maybe                           (Maybe (Just, Nothing), fromMaybe)
import           Data.Text.Lazy                       (Text, pack)
import           Foreign.Marshal.Unsafe               (unsafeLocalState)
import           Network.HTTP.Types.Status            (methodNotAllowed405)
import           Network.Wai.Middleware.RequestLogger (logStdout)
import           Relude                               (Eq ((/=)), FilePath, Foldable (length, null), IO, Int,
                                                       Monad (return, (>>), (>>=)), Num ((+)), Ord ((<)), String, const,
                                                       exitFailure, intercalate, map, otherwise, putStrLn, readFile,
                                                       seq, show, stderr, sum, unless, when, ($), (++), (.), (=<<))
import           System.Directory                     (doesFileExist)
import           System.Environment.MrEnv             (envAsInt, envAsString)
import           System.IO                            (writeFile)
import           System.IO.Strict                     (readFile)
import           Web.Scotty                           (body, delete, get, json, middleware, notFound, param, post,
                                                       redirect, scotty, status, text)

-- -----------------------------------------------------------------------------

loadData :: FilePath -> Map String Int
loadData db = fromMaybe Data.Map.empty $ unsafeLocalState $ decodeFileStrict' db
{-# NOINLINE loadData #-}

total :: FilePath -> Map String Int
total db = Data.Map.singleton "total" $ Relude.sum $ loadData db
{-# NOINLINE total #-}

normalize :: String -> String
normalize (x:xs) = Data.Char.toUpper x : Relude.map Data.Char.toLower xs
normalize []     = []

getAll :: FilePath -> Map String Int
getAll db = fromMaybe Data.Map.empty $ decode' $ unsafeLocalState $ do
        contents <- Relude.readFile db
        seq (Relude.length contents) (return ())

        return $ BL.pack contents
{-# NOINLINE getAll #-}

getData :: FilePath -> String -> Map String Int
getData db x = Data.Map.singleton (normalize x) $ findWithDefault 0 (normalize x) (loadData db)
{-# NOINLINE getData #-}

updateData :: FilePath -> BL.ByteString -> String -> Map String Int
updateData db val key = unsafeLocalState $ do
        q <- System.IO.Strict.readFile db

        let js :: Maybe (Map String Int)
            js = decode' $ BL.pack q in
                let res = j $ fromMaybe Data.Map.empty js in do
                    Data.Aeson.encodeFile db res
                    return res
    where
        j :: Map String Int -> Map String Int
        j db = Data.Map.filter (/= 0) $ insertWith fx (normalize key) int db

        int :: Int
        int = m $ BL.readInt val

        fx :: Int -> Int -> Int
        fx = case BL.head val of
            '+' -> (+)
            '-' -> (+)
            _   -> const

        m :: Maybe (Int, BL.ByteString) -> Int
        m (Just (x, _)) = x
        m Nothing       = 0

deleteData :: FilePath -> String -> Map String Int
deleteData db k = unsafeLocalState $ do
        q <- System.IO.Strict.readFile db

        let js :: Maybe (Map String Int)
            js = decode' $ BL.pack q in
                let res = j $ fromMaybe Data.Map.empty js in do
                    Data.Aeson.encodeFile db res
                    return res
    where
        j :: Map String Int -> Map String Int
        j d = Data.Map.filter (/= 0) $ Data.Map.delete (normalize k) d

metrics :: String -> Data.Text.Lazy.Text
metrics = Data.Text.Lazy.pack . Relude.intercalate "\\n" . elems . Data.Map.mapWithKey m . loadData
    where
        m :: String -> Int -> String
        m k v = "finex_amount_" ++ t v ++ "{type=\"" ++ k ++"\"} " ++ show v

        t :: Int -> String
        t x | x < 0     = "savings"
            | otherwise = "debts"


startServer :: Int -> String -> IO ()
startServer port db = scotty port $ do
        middleware logStdout

        Web.Scotty.get    "/"               $ Web.Scotty.json $ Main.getAll db
        Web.Scotty.post   "/"               $ status methodNotAllowed405
        Web.Scotty.delete "/"               $ status methodNotAllowed405

        Web.Scotty.get    "/metrics"        $ Web.Scotty.text $ metrics db
        Web.Scotty.post   "/metrics"        $ status methodNotAllowed405
        Web.Scotty.delete "/metrics"        $ status methodNotAllowed405

        Web.Scotty.get    "/total"          $ Web.Scotty.json $ total db
        Web.Scotty.post   "/total"          $ status methodNotAllowed405
        Web.Scotty.delete "/total"          $ status methodNotAllowed405

        Web.Scotty.get    "/:path"          $ Web.Scotty.json . getData db =<< param "path"
        Web.Scotty.post   "/:path"          $ body >>= \b -> Web.Scotty.json . updateData db b =<< param "path"
        Web.Scotty.delete "/:path"          $ Web.Scotty.json . deleteData db =<< param "path"

        notFound                            $ redirect "/"

main :: IO ()
main = do
        port <- envAsInt    "FINEX_PORT" 3000
        db   <- envAsString "FINEX_DB"   []

        fileExists <- doesFileExist db

        when (Relude.null db) $ BS.hPutStrLn stderr "ERROR: FINEX_DB environment variable is not set." >> Relude.exitFailure

        unless fileExists $ System.IO.writeFile db "" >> Relude.putStrLn ("Database file " ++ db ++ " does not exist. Creating.")

        startServer port db

