#lang racket/base

;; Copyright (C) 2022, Maxim Likhachev, <envrm@yandex.ru>
;;
;; This program is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation, version 3.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program. If not, see <http://www.gnu.org/licenses/>.

(require response-ext
         json
         racket/file
         racket/string
         routy
         web-server/servlet
         web-server/servlet-env)

(define (load-data)
  (if (file-exists? db)
    (file->string db)
    "{}"))

(define (load-data-json)
  (string->jsexpr (load-data)))

(define (save-data data)
  (display-to-file data db #:exists 'replace))

(define (sum-values data)
  (jsexpr->string (hash 'total
                        (foldl + 0 (hash-values data)))))

(define (metrics)
  (string-join
    (append (list "# HELP finex_amount_debts Долги"
              "# TYPE finex_amount_debts gauge"
              "# HELP finex_amount_savings Сбережения"
              "# TYPE finex_amount_savings gauge"
              (format "finex_memory_usage ~a" (current-memory-use)))
            (hash-map (load-data-json)
                      (λ (k v)
                        (format "finex_amount_~a{type=\"~a\"} ~a"
                                (if (positive? v) 'savings 'debts) k v)))) "\n"))

(define (amount/get key)
  (let* ((data (load-data-json))
         (f (string->symbol (string-titlecase key))))

    (jsexpr->string (json-null))))

(define (amount/update key value)
  (if value
    (let* ((data (load-data-json))
           (f (string->symbol (string-titlecase key)))
           (number-value (string->number (bytes->string/utf-8 value)))
           (current-value (hash-ref data f 0))
           (new-value (if (not number-value) #f
                        (case (bytes-utf-8-ref value 0)
                          ((#\+ #\-) (+ current-value number-value))
                          (else number-value))))
           (new-data (case new-value
                       ((#f) (raise (error (format "Invalid number: ~a" value))))
                       ((0) (jsexpr->string (hash-remove data f)))
                       (else (jsexpr->string (hash-set data f new-value))))))
      (save-data new-data)
      new-data)
    (raise (error "Cannot update key new value is not set"))))

(define (amount/delete key)
  (let* ((data (load-data-json))
         (f (string->symbol (string-titlecase key)))
         (new-data (jsexpr->string (hash-remove data f))))
    (save-data new-data)
    new-data))

(define (response/405)
  (response/make #:code 405 #:message (string->bytes/utf-8 "Method Not Allowed") ""))

; ------------------------------------------------------------------------------

(routy/get       "."        (λ (req params) (load-data)))
(routy/get       "/all"     (λ (req params) (load-data)))
(routy/get       "/total"   (λ (req params) (sum-values (load-data-json))))
(routy/get       "/metrics" (λ (req params)  (metrics)))

(routy/post      "."        (λ (req params) (response/405)))
(routy/post      "/all"     (λ (req params) (response/405)))
(routy/post      "/total"   (λ (req params) (response/405)))
(routy/post      "/metrics" (λ (req params) (response/405)))

(routy/delete    "."        (λ (req params) (response/405)))
(routy/delete    "/all"     (λ (req params) (response/405)))
(routy/delete    "/total"   (λ (req params) (response/405)))
(routy/delete    "/metrics" (λ (req params) (response/405)))

(routy/get       "/:path"   (λ (req params) (amount/get (request/param params 'path))))
(routy/post      "/:path"   (λ (req params) (amount/update (request/param params 'path) (request-post-data/raw req))))
(routy/delete    "/:path"   (λ (req params) (amount/delete (request/param params 'path))))

(routy/not-found            (λ (req) (load-data)))

; ------------------------------------------------------------------------------

(define db   (getenv "FINEX_DB"))
(define host (getenv "FINEX_HOST"))
(define port (getenv "FINEX_PORT"))

(unless db
  (eprintf "ERROR: FINEX_DB environment variable is not set.~%")
  (exit 1))

(unless (non-empty-string? host)
 (displayln "FINEX_HOST is not set. Using 127.0.0.1."))

(unless (non-empty-string? port)
 (displayln "FINEX_PORT is not set. Using 8000."))

(serve/servlet
  (λ (req) (routy/response req))
  #:launch-browser? #f
  #:servlet-path "/"
  #:listen-ip (if host host "127.0.0.1")
  #:port (if port (string->number port) 8000)
  #:log-file (current-output-port)
  #:servlet-regexp #rx"")

